/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

 /**
 * @file cc/crc.c
 *
 * @brief CRC-16 CCITT routine
  *
 * Polynomial: x^16 + x^12 + x^5 + 1 (0x8408)
 * Initial value: 0xffff
 *
 * Needed only for unix emulation as crc_ccitt_update is included in avr-libc
 *
 * 2004/06/19 Matthias Ringwald <mringwal@inf.ethz.ch>
 *
 */

#include <sys/types.h>

#ifdef __AVR_ARCH__

#if __AVR_LIBC_VERSION__ >= 10400UL
#include <util/crc16.h>
#else
#include <avr/crc16.h>
#endif

u_short crc_ccitt_update(u_short crc, u_char data)
{
    return _crc_ccitt_update(crc, data);
}

#else

u_short crc_ccitt_update(u_short crc, u_char data)
{
    data ^= crc & 0xff;
    data ^= data << 4;

    return ((((u_short) data << 8) | (crc >> 8)) ^ (u_short) (data >> 4) ^ ((u_short) data << 3));
}

#endif

/** 

\brief Convenience function to compute the CRC-16 CCITT of a data buffer

\param buf pointer to data crc shoud be computed
\param size of data buffer
\return crc-16 value

 */

u_short crc_ccitt_compute(u_char * buf, u_short size)
{
    u_short index;

    // initial value for CRC-16 CCITT
    u_short crc = 0xffff;

    for (index = 0; index < size; index++) {
        crc = crc_ccitt_update(crc, buf[index]);
    }
    return crc;
}
