/*!
 * \file con_mgr.c
 * \brief Implementation of the Connection Manager
 * 
 * This is the simplified version of the connection manager.
 * See singleroot_conmgr.c for an alternative version.
 * 
 * The connection manager builds and
 * maintains a connected network in a distributed fashion.
 * 
 * For now, a BTnode that has reached its maximum number of master and slave
 * roles does not participate in connection-building anymore.
 * 
 * The connection manager consists of two threads:
 * 
 * One thread performs inquiries and connects to
 * neighboring nodes with the same class of device. The thread waits
 * rand()/RAND_MAX times sleep_period seconds between inquiries.
 * 
 * A second thread reacts to events (connect, disconnect) and handles incoming
 * packets (tree-id, negotiation, negotiation-response).
 * 
 * When two nodes connect, the node with a lower tree ID takes over the other
 * node's tree ID and broadcasts it to its former tree. Like this, the new tree
 * ID is distributed to all nodes in the new tree.
 * 
 * A node also sends (i.e. initiates) such a broadcast if
 * - it received a tree-ID broadcast with a lower ID than its own MAC address:
 * the node's MAC address is broadcast.
 * - one of its connections was unvoluntarily dropped: the node's MAC address is
 * broadcast.
 * A node forwards a received broadcast if and only if
 * - the tree-ID in the broadcast is higher than its own MAC address
 * - the broadcast message was not previously received from another link; in
 * this case, the link over which the message just arrived is dropped. This
 * removes cycles from the graph and thus maintains a tree structure.
 * 
 * The master and slave roles for the connection are decided based on balancing
 * criteria.
 * 
 * Further ideas:
 * - A node inquires less if its tree ID has been raised many times (this
 * means that many connections have already been established). If the ID is
 * lowered, there has been a disconnect and the node inquires more frequently.
 * - Use class-of-device bits to give information about available master/slave
 * roles and the tree ID; these bits are transferred during inquiry at no cost!
 * 
 * 
 */

#include <led/btn-led.h>
#include <sys/timer.h>
#include <stdlib.h>
#include <bt/bt_hci_api.h>
#include <bt/bt_hci_cmds.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_defs.h>
#include <terminal/btn-terminal.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/thread.h>
#include <sys/heap.h>
#include <sys/mutex.h>
#include <sys/semaphore.h>
#include <eepromdb/btn-eepromdb.h>
#include <string.h>
#include <bt/bt_psm.h>
#include <bt/l2cap_cl.h>
#include <debug/logging.h>
#include <debug/toolbox.h>
#include <cm/con_event_buffer.h>
#include <cm/_con_mgr.h>

// debug settings
#define LOG_LEVEL 0
#define LOG_CLASS 5
#include <debug/log_set.h>

// maximum slave roles that a node is willing to take
#define CM_MAX_SLAVE_ROLES            1
#define CM_MAX_SLAVES                 6
#define CM_AUTINQ          CM_AUTINQ_ON
#define CM_AIPERIOD                  20
#define CM_AITIME                     3
#define CM_MAX_INQ_DEVICES           10

//pkt types
#define CM_PKT_TYPE_TREE_ID           0
#define CM_PKT_TYPE_NEGOT             1
#define CM_PKT_TYPE_NEGOT_RESP        2


//heartbeat (only showing if device is not connected)
#define CM_LED_HB_PAUSE              50
#define CM_LED_HB_DURATION            2

// minimum sleep time between inquiries [seconds]
#define CM_MIN_SLEEP                  5

//autinq vals
#define CM_AUTINQ_OFF                 0
#define CM_AUTINQ_ON                  1

 ///< link supervision timeout (1600 slots = 1s)
 #define CM_LINK_TIMEOUT           11200


typedef struct _cm_stack {
    bt_addr_t tree_ID;
    bt_addr_t my_addr;
    bt_hci_con_handle_t tree_id_handle;
    struct bt_hci_cmd_response response;
    bt_hci_con_handle_t reliable_cons[BT_HCI_MAX_NUM_CON];
    struct btstack *bt_stack;   // pointer to lower level hci stack 
    bt_psm_t* psmux;
    u_short psm;
    bt_acl_pkt_buf* buf_que;		///< Pointer to the packet buffers
    u_char aut_inquiry;
    u_char inquiries_since_discon;
    u_char num_of_reliable_cons;
    u_char sleep_period;
    u_char ai_time;
    SEM inq_mutex;
    u_char data_buf[115];
//    char bt_addr_buf[18];
    MUTEX rel_con_cb_mutex;
    u_char max_slaves;  ///< maximum number of slaves that this node is willing to accept
    u_long cod;
} cm_stack_t;

cm_stack_t* cm_stack;

// read autinq from EEPROM and set it; if EEPROM is empty, use val
u_char cm_tree_set_autinq_to_EEPROM(u_char val) {
		unsigned int len, read_EEPROM = 0;

        len = btn_eepromdb_get(FOURCC('A', 'I', 'N', '='), sizeof(unsigned int), &read_EEPROM);
        if (len == 0) {
          cm_stack->aut_inquiry = val;
        }
        else {
          cm_stack->aut_inquiry = read_EEPROM;
        }
        return cm_stack->aut_inquiry;
}

void cm_tree_set_autinq(u_char val) {
        cm_stack->aut_inquiry = val;
}

void cm_tree_set_inquiries(u_char val) {
  cm_stack->inquiries_since_discon = val;
}

void _bt_cm_set_sleep_period(u_char val) {
  cm_stack->sleep_period = val;
}

void _bt_cm_print_reliable_cons(void)
{
    u_char i;
    tprintf("reliable connections:\n");
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if (cm_stack->reliable_cons[i] != BT_HCI_HANDLE_INVALID){
            tprintf("%d, ", cm_stack->reliable_cons[i]);
        }
    }
    tprintf("\n");
}

void _bt_cm_print_tree_ID(void)
{
    tprintf("Local tree ID: "ADDR_FMT"\n", ADDR(cm_stack->tree_ID));
}

u_char _bt_cm_get_nr_reliable_cons(void)
{
    u_char nr_open_con = 0, i;
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if (cm_stack->reliable_cons[i] != BT_HCI_HANDLE_INVALID)
            nr_open_con++;
    }
    return nr_open_con;
}

void _bt_cm_disconnect_all(void)
{
    u_char i;
    //disconn all
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if (cm_stack->reliable_cons[i] != BT_HCI_HANDLE_INVALID) {
            DEBUG("discon handle %d\n", cm_stack->reliable_cons[i]);
            if(bt_hci_disconnect(cm_stack->bt_stack,
                               BT_HCI_SYNC,
                               cm_stack->reliable_cons[i],
                               BT_HCI_DISCONNECT_USER_ENDED)) {
                ERROR("can not disconnect handle %d\n", cm_stack->reliable_cons[i]);
            }
            DEBUG("done\n");
        }
    }
}


/* The following function is for sending a broadcast packet (containing a new tree ID)
 */
void _bt_cm_broadcast(u_char* data, bt_hci_con_handle_t except_con_handle, u_short len)
{
    u_char i;
    // send packet to all reliable connections
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if ((cm_stack->reliable_cons[i] != BT_HCI_HANDLE_INVALID)
             && (cm_stack->reliable_cons[i] != except_con_handle)) { // incoming 
            if(l2cap_cl_send(data, len, cm_stack->reliable_cons[i], cm_stack->psm)) {
                ERROR("cannot send to device with handle %d\n", cm_stack->reliable_cons[i]);
            }
        }
    }
}

u_char _cm_accept_filter(bt_addr_t addr, u_long cod, u_char type, u_char *myrole, void* arg)
{
    cm_stack_t* cm_stack = (cm_stack_t*)arg;
    *myrole = BT_HCI_MY_ROLE_SLAVE;
    if(cod == cm_stack->cod){
        return 1;    
    } 
    return 0;
}


/*
 * checks if the connection with this handle is a reliable connection
 */
u_char _bt_cm_is_reliable_con( bt_hci_con_handle_t con_handle)
{
    u_char i;
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if (cm_stack->reliable_cons[i] == con_handle) {
            break;
        }
    }
    return (i != BT_HCI_MAX_NUM_CON);
}

void _bt_cm_add_reliable_con( bt_hci_con_handle_t con_handle)
{
    u_char i;
    DEBUG("adding con\n");
    if (cm_stack->num_of_reliable_cons == 0) {
     
        //first connection
        btn_led_heartbeat(0, 0, 0);
    }
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if (cm_stack->reliable_cons[i] == con_handle)
            return;
        if (cm_stack->reliable_cons[i] == BT_HCI_HANDLE_INVALID) {
            cm_stack->reliable_cons[i] = con_handle;
            break;
        }
    }
    cm_stack->num_of_reliable_cons++;
    //INFO("writing new cod\n");
    //bt_hci_write_local_cod(cm_stack->bt_stack, BT_HCI_SYNC, CM_INIT_COD + cm_stack->num_of_reliable_cons);
    //turn off heartbeat by passing 0,0,0
    INFO("reliable connection (handle %d) added\n", con_handle);
}

void _bt_cm_remove_reliable_con( bt_hci_con_handle_t con_handle)
{
    u_char i;

    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if (cm_stack->reliable_cons[i] == con_handle) {
            cm_stack->reliable_cons[i] = BT_HCI_HANDLE_INVALID;
            cm_stack->num_of_reliable_cons--;
            break;
        }
    }
    if (i == BT_HCI_MAX_NUM_CON)
        return;
    if (cm_stack->num_of_reliable_cons == 0) {
        btn_led_heartbeat(CM_LED_HB_PAUSE, 1, CM_LED_HB_DURATION);
    }
    //bt_hci_write_local_cod(cm_stack->bt_stack, BT_HCI_SYNC, CM_INIT_COD + cm_stack->num_of_reliable_cons);
}

u_char _bt_cm_lower_bt_addr(bt_addr_t id1, bt_addr_t id2)
{
    int i;
    u_char retvalue = 0;
    for (i = BD_ADDR_LEN - 1; i >= 0; i--) {
        if (id1[i] < id2[i]) {
            retvalue = 1;
            break;
        } else if (id1[i] > id2[i]) {
            retvalue = 0;
            break;
        }
    }
    return retvalue;
}

void _bt_cm_show_tree_ID(void)
{
    if (cm_stack->tree_ID[0] & 0x1)
        btn_led_set(0);
    else
        btn_led_clear(0);
    if (cm_stack->tree_ID[0] & 0x2)
        btn_led_set(1);
    else
        btn_led_clear(1);
    if (cm_stack->tree_ID[0] & 0x4)
        btn_led_set(2);
    else
        btn_led_clear(2);
    if (cm_stack->tree_ID[0] & 0x8)
        btn_led_set(3);
    else
        btn_led_clear(3);
}

/*!
 * \brief processes a received tree id.
 * \param[in] received_tree_ID
 * \param[in] con_handle
 * \param[in] type
 * \return 0 on success, 1 if received_tree_ID matches our own tree id.
 */
u_char _bt_cm_process_tree_ID(bt_addr_t received_tree_ID, bt_hci_con_handle_t con_handle, char type)
{
    u_char *payload;

    if (_bt_cm_lower_bt_addr(cm_stack->tree_ID, received_tree_ID)) {
        // a higher ID was received, copy received_tree_ID to tree_ID
        INFO("_bt_cm_process_tree_ID: a higher tree ID ("ADDR_FMT") " \
                "was received (handle %d)\n",
                ADDR(received_tree_ID),
                con_handle);
        memcpy(cm_stack->tree_ID, received_tree_ID, BD_ADDR_LEN);        // overwrite my old tree ID
        _bt_cm_show_tree_ID();
        cm_stack->tree_id_handle = con_handle;
        payload = cm_stack->data_buf;
        payload[0] = cm_stack->psm;
        payload[1] = CM_PKT_TYPE_TREE_ID;       // set packet type
        memcpy(payload + 2, cm_stack->tree_ID, BD_ADDR_LEN);
        _bt_cm_broadcast(payload, con_handle, BD_ADDR_LEN + 2);
        return 0;
    } else if (_bt_cm_lower_bt_addr(received_tree_ID, cm_stack->tree_ID)) {
        // a lower ID was received
        INFO("_bt_cm_process_tree_ID: a lower tree ID ("ADDR_FMT") " \
                "was received (handle %d)\n",
                ADDR(received_tree_ID),
                con_handle);
        if (con_handle == cm_stack->tree_id_handle) {    // a connection was lost
            memcpy(cm_stack->tree_ID, received_tree_ID, BD_ADDR_LEN);
            _bt_cm_show_tree_ID();
            payload = cm_stack->data_buf;
            payload[0] = cm_stack->psm;
            payload[1] = CM_PKT_TYPE_TREE_ID;   // set packet type
            memcpy(payload + 2, cm_stack->tree_ID, BD_ADDR_LEN);
            _bt_cm_broadcast(payload, con_handle, BD_ADDR_LEN + 2);
        }
        return 0;
    } else {
        /* received_tree_ID == tree_ID, i.e. there's a cycle */
        if (type != CM_PKT_TYPE_NEGOT_RESP) {
            INFO("_bt_cm_process_tree_ID: identical tree ID ("ADDR_FMT") " \
                    "was received, disconnecting (handle %d)\n",
                    ADDR(received_tree_ID),
                    con_handle);
            /* TODO add error check */
            bt_hci_disconnect(cm_stack->bt_stack, BT_HCI_SYNC, con_handle, BT_HCI_DISCONNECT_USER_ENDED);
            return 1;
        }
        return 0;
    }

}

void _change_con_pkt_type(bt_hci_con_handle_t con_handle)
{
    long res;
    DEBUG("changing conpkt type\n");
    res = bt_hci_change_con_pkt_type(cm_stack->bt_stack, BT_HCI_SYNC, con_handle, BT_HCI_PACKET_TYPE_DM3);
    if (res < 0) {
        ERROR("con pkt change: error\n");
    }
}

void _change_link_policy(bt_hci_con_handle_t con_handle)
{
    long res;
    res = bt_hci_write_link_policy_settings(cm_stack->bt_stack, BT_HCI_SYNC, con_handle, BT_HCI_LINK_POLICY_ROLE_SWITCH);
    if (res < 0) {
        ERROR("link policy: error\n");
    }
}

void _change_disconnect_timeout( bt_hci_con_handle_t handle)
{
    long res;
    bt_hci_con_handle_t con_handle;
    con_handle = handle;
    // TODO add error check
    //bt_l2cap_get_hci_connection_handle( handle, &con_handle);
    DEBUG("changing disc timeout\n");
    res = bt_hci_write_link_supervision_timeout(cm_stack->bt_stack, BT_HCI_SYNC, con_handle, CM_LINK_TIMEOUT);
    if (res < 0) {
        ERROR("error changing disc timeout\n");
    }
}

u_char _bt_cm_count_reliable_roles(u_char role)
{
    u_char i, right_role = 0;
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if (cm_stack->reliable_cons[i] != BT_HCI_HANDLE_INVALID) {
            bt_hci_con_handle_t con_handle;
            con_handle = cm_stack->reliable_cons[i];
            if (bt_hci_local_role_discovery(cm_stack->bt_stack, con_handle) == role) {
                right_role++;
            }
        }
    }
    return right_role;
}

/* Here comes the code for handling changes in the connection table.
 */
void _con_table_change_cb(u_char type, u_char detail, bt_hci_con_handle_t con_handle, void* cb_arg)
{
    u_char *data;
    // change connection properties
    if (type == BT_HCI_CONNECTION) {
        if (detail == BT_HCI_CONNECTION_ACTIVE){
            _change_disconnect_timeout(con_handle);
        } else {
        }
        _change_con_pkt_type(con_handle);
        //change_link_policy(con_handle); //for rolechange
    }

    switch (type) {
    case BT_HCI_DISCONNECTION:
        /* if the link was lost (i.e. no voluntary disconnect) and we
         * received our tree ID over this link, we have to reset our
         * tree ID to our BT MAC address.
         */
        if (detail == BT_HCI_DISCON_LINK_LOST) {
           cm_stack->inquiries_since_discon=0;
        } else {
        }
        INFO("DISCONNECTION (handle %d)\n", con_handle);
        NutMutexLock(&cm_stack->rel_con_cb_mutex);

        if (_bt_cm_is_reliable_con(con_handle)) {
            con_mgr_stack.con_table_cb(type, detail, con_handle, con_mgr_stack.con_table_cb_arg);
            _bt_cm_remove_reliable_con(con_handle);            
        }
        if (con_handle == cm_stack->tree_id_handle) {
            INFO("ID-giving link lost\n");
            // overwrite tree ID with my MAC
            memcpy(cm_stack->tree_ID, cm_stack->my_addr, BD_ADDR_LEN);
            _bt_cm_show_tree_ID();
            cm_stack->tree_id_handle = BT_HCI_HANDLE_INVALID;
            data = cm_stack->data_buf;
            data[0] = cm_stack->psm;
            data[1] = CM_PKT_TYPE_TREE_ID;   // set packet type
            memcpy(data + 2, cm_stack->tree_ID, BD_ADDR_LEN);
            _bt_cm_broadcast(data, BT_HCI_HANDLE_INVALID, BD_ADDR_LEN + 2);
        }
        NutMutexUnlock(&cm_stack->rel_con_cb_mutex);

        break;
    case BT_HCI_CONNECTION:
        /* A connection was just established; if we initiated it, then we
         * send a negotiation packet. Otherwise the
         * connection-manager thread will take care of the incoming
         * negotiation packet and perform all further steps.
         */
        INFO("CONNECTION (handle %d)\n", con_handle);
        if (detail == BT_HCI_CONNECTION_ACTIVE) {
            data = cm_stack->data_buf;
            data[0] = cm_stack->psm;
            data[1] = CM_PKT_TYPE_NEGOT;     // set packet type
            memcpy(data + 2, cm_stack->tree_ID, BD_ADDR_LEN);
            data[BD_ADDR_LEN + 2] = _bt_cm_count_reliable_roles(BT_HCI_MY_ROLE_MASTER);
            INFO("con_table_change_cb: sending negotiation packet (type %d) (handle %d)\n", payload[1], con_handle);
            if(l2cap_cl_send(data, BD_ADDR_LEN + 3, con_handle, cm_stack->psm)) {
                ERROR("could not send neg pkt, disconnecting.\n");
                bt_hci_disconnect(cm_stack->bt_stack, BT_HCI_SYNC, con_handle, BT_HCI_DISCONNECT_USER_ENDED);
            }
        }
        break;
    default:
        WARNING("other type than con/discon\n!");
    }
    DEBUG("con_table_change_cb: done\n");
}



/*
 * data callback function
 */
bt_msg* _cm_data_cb(bt_msg* msg,
					u_char* data,
					u_short len,
					u_short service_nr,
					void* cb_arg)
{
	bt_hci_con_handle_t con_handle;
    bt_addr_t received_ID;
//      struct bt_hci_cmd_response response;
    u_char my_slave_roles, received_slave_roles;
    //variables for checking the bt-stack-con-table
    //char get_packet_return;
    //bt_hci_con_handle_array con_handles;
    //u_char i, j, found, nr_open_con;

    //changed to check the bt-stack-con-table with the reliable connections and signal an error
    /*
    get_packet_return = BT_ERR_L2CAP_TIMEOUT;
    while (get_packet_return == BT_ERR_L2CAP_TIMEOUT) {
        get_packet_return = bt_acl_com_get_packet(CM_ACL_COM_SERVICE_NR, &acl_pkt_p, &con_handle, &len, 10000);

        nr_open_con = (u_char) bt_hci_get_con_handles(cm_stack->bt_stack, con_handles);
        for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
            if (cm_stack->reliable_cons[i] != BT_HCI_HANDLE_INVALID) {
                found = 0;
                for (j = 0; j < nr_open_con; j++) {
                    if (con_handles[j] == cm_stack->reliable_cons[i]) {
                        found = 1;
                        break;
                    }
                }
                if (!found) {
                    //ERROR, the disconnection callback is missing to this change or
                    //it will arrive any minute :)
                    DEBUG("CM: found con handle the is not in bt-stack\n");
                    con_table_change_cb(BT_HCI_DISCONNECTION, BT_HCI_DISCON_LINK_LOST, cm_stack->reliable_cons[i]);
                }
            }
        }
    }
    if (get_packet_return) {
        WARNING("bt_acl_com_get_packet returned error\n");
        continue;
    }*/

	// get con handle
	con_handle = bt_acl_get_con_handle(msg->pkt);
	
    switch (data[1]) {   // type
        //------------------------------------------------------------------
    case CM_PKT_TYPE_TREE_ID:
        DEBUG("received tree-ID packet (handle %d)\n", con_handle);
        memcpy(received_ID, &data[2], BD_ADDR_LEN);
        _bt_cm_process_tree_ID(received_ID, con_handle, CM_PKT_TYPE_TREE_ID);
        break;
        //------------------------------------------------------------------
    case CM_PKT_TYPE_NEGOT:
        // we assume that this node is the slave, this might change
        // with BTnode rev3!
        DEBUG("CM: received neg. packet (handle %d)\n", con_handle);
        
	    NutMutexLock(&cm_stack->rel_con_cb_mutex);
        DEBUG("CM:thread enter mutex\n");
        
        my_slave_roles = _bt_cm_count_reliable_roles(BT_HCI_MY_ROLE_MASTER);
//                              if (my_slave_roles <= CM_MAX_SLAVE_ROLES) {
//                      //DEBUG("CM: one slave role still available (handle %d)\n", con_handle);
        memcpy(received_ID, &data[2], BD_ADDR_LEN);
        received_slave_roles = data[BD_ADDR_LEN + 2];
//                                      // now decide who is master and who is slave
//                                      if (my_slave_roles < received_slave_roles || (my_slave_roles == received_slave_roles && _bt_cm_lower_bt_addr(tree_ID, received_ID))) {
//                              DEBUG("CM: I should be slave (handle %d)\n", con_handle);
//                                              // this node should be the slave, a role switch is maybe necessary
//                                              if (bt_hci_local_role_discovery(_bt_conmgr_bt_stack, con_handle) == BT_HCI_MY_ROLE_MASTER) {
//                                      //      DEBUG("CM: initiating role switch (handle %d)\n", con_handle);
//                                              //      bt_hci_role_set(_bt_conmgr_bt_stack, &response, con_handle, BT_HCI_MY_ROLE_SLAVE);
//                                              }
//                                      }
        
        // TODO why?
        // check if this con handle is still in stack con table or it connection just disconnected!
         if (bt_hci_is_con_open(cm_stack->bt_stack, con_handle) != 0) {
            ERROR("con handle not in stack anymore!!\n");
            NutMutexUnlock(&cm_stack->rel_con_cb_mutex);
            DEBUG("thread exit mutex\n");
            break;
         }
        
        DEBUG("processing tree ID (handle %d)\n", con_handle);
            
        if( _bt_cm_process_tree_ID( received_ID,
                                    con_handle,
                                    CM_PKT_TYPE_NEGOT) == 0) {
            _bt_cm_add_reliable_con(con_handle);
            DEBUG("call jaws\n");
            con_mgr_stack.con_table_cb(BT_HCI_CONNECTION,
            							BT_HCI_CONNECTION_PASSIVE,
            							con_handle,
            							con_mgr_stack.con_table_cb_arg);
            data = cm_stack->data_buf;
            data[0] = cm_stack->psm;
            data[1] = CM_PKT_TYPE_NEGOT_RESP;        // set packet type
            memcpy( data + 2, cm_stack->tree_ID, BD_ADDR_LEN);
            if (l2cap_cl_send(data, BD_ADDR_LEN, con_handle, cm_stack->psm)) {
                ERROR("could not send neg-resp pkt, disconnecting.\n");
                // TODO add error check
                bt_hci_disconnect(cm_stack->bt_stack, BT_HCI_SYNC, con_handle, BT_HCI_DISCONNECT_USER_ENDED);
                NutMutexUnlock(&cm_stack->rel_con_cb_mutex);
                DEBUG("thread exit mutex\n");
                break;
               
            }
        }

//                              }
//                              else {
//                      DEBUG("CM: no slave role available, disconnecting (handle %d)\n", con_handle);
//                                      bt_hci_disconnect(_bt_conmgr_bt_stack, &response, con_handle, BT_HCI_DISCONNECT_USER_ENDED);                    
//                              }
	    NutMutexUnlock(&cm_stack->rel_con_cb_mutex);
        DEBUG("CM:thread exit mutex\n");
        break;
        //------------------------------------------------------------------
    case CM_PKT_TYPE_NEGOT_RESP:
        
        DEBUG("CM: received neg.-response pkt (handle %d)\n", con_handle);
	    NutMutexLock(&cm_stack->rel_con_cb_mutex);
        DEBUG("CM:thread enter mutex\n");
       
        memcpy(received_ID, data + 2, BD_ADDR_LEN);
        // TODO why?
        // check if this con handle is still in stack con table or it connection just disconnected!
        if (bt_hci_is_con_open(cm_stack->bt_stack, con_handle) != 0) {
            ERROR("con handle not in stack anymore!!\n");
            NutMutexUnlock(&cm_stack->rel_con_cb_mutex);
            DEBUG("CM:thread exit mutex\n");
            break;
        }
        DEBUG("processing tree ID (handle %d)\n", con_handle);
        if( _bt_cm_process_tree_ID( received_ID,
                                    con_handle,
                                    CM_PKT_TYPE_NEGOT_RESP) == 0) {
            _bt_cm_add_reliable_con(con_handle);
            con_mgr_stack.con_table_cb(BT_HCI_CONNECTION,
            							BT_HCI_CONNECTION_ACTIVE,
            							con_handle,
            							con_mgr_stack.con_table_cb_arg);
        }
        NutMutexUnlock(&cm_stack->rel_con_cb_mutex);
        DEBUG("thread exit mutex\n");
        break;
        //------------------------------------------------------------------
    default:
        WARNING("unknown CM packet type %d\n", data[1]);
        break;
        //------------------------------------------------------------------
    }// end switch
    return msg;
}

/*
 * inquiry thread
 * This thread is responsible for doing regular inquiries and connecting to one
 * of the compatible devices (CoD). The behavior can be controlled from the
 * command line with the "cm autinq" and "cm aiperiod" commands.
 */
char CONMGR_INQUIRY_T_bt_addr_buf[18];
THREAD(CONMGR_INQUIRY_T, arg)
{
    int sleep_period;
	// int num_of_inquiries = 0;
    long connect_result;
    struct bt_hci_inquiry_result _bt_cmds_inqres[CM_MAX_INQ_DEVICES];
    int _right_cod_dev_indices[CM_MAX_INQ_DEVICES];
    int _bt_cmds_nr_inq_dev = 0, _right_cod_dev = 0;
    int _chosen_device, i, j;
//    u_char _bt_cmds_inq_time = 5;
    u_long max_CoD = 0;
    bt_hci_con_handle_array con_handles;
    u_char nr_open_con;
    bt_addr_t addr;

    NutThreadSetPriority(100);
    NutSleep(2000);
    for (;;) {
        while ((cm_stack->aut_inquiry == CM_AUTINQ_ON) && 
        (_bt_cm_count_reliable_roles(BT_HCI_MY_ROLE_MASTER) < cm_stack->max_slaves)) {
            _right_cod_dev = 0;
            while (!_right_cod_dev) {
                // make us visible for inquiries and connects
                //bt_hci_write_scan_enable(cm_stack->bt_stack, BT_HCI_SYNC, BT_HCI_SCAN_MODE_INQUIRY_PAGE);
                
                sleep_period = CM_MIN_SLEEP + ((1 + (++(cm_stack->inquiries_since_discon)))/3)*(int)(((float)cm_stack->sleep_period) * rand() / (RAND_MAX + 1.0));
                DEBUG("inq sleep %d s\n", sleep_period);
                NutSleep(1000 * (u_long)sleep_period);
                //check if automatic inquiry should be done
                if (cm_stack->aut_inquiry == CM_AUTINQ_OFF) {
                    cm_stack->inquiries_since_discon = 0;
                    break;
                }
                               
                DEBUG("inq\n");
                // wait for inquiry mutex 
                NutSemWait(&cm_stack->inq_mutex);
                DEBUG("after inq mutex\n");
				// make us invisible for inquiries and connects
                //bt_hci_write_scan_enable(cm_stack->bt_stack, BT_HCI_SYNC, BT_HCI_SCAN_MODE_OFF);
                _bt_cmds_nr_inq_dev =
                    bt_hci_inquiry(cm_stack->bt_stack, BT_HCI_SYNC, cm_stack->ai_time, CM_MAX_INQ_DEVICES, _bt_cmds_inqres);
                NutSemPost(&cm_stack->inq_mutex);
                btn_led_add_pattern(BTN_LED_PATTERN_KNIGHT, 0, 1, 1);
                DEBUG("inq done: %d devices found.\n", _bt_cmds_nr_inq_dev);
                
                // Now select the found devices with the right class of device
                _right_cod_dev = 0;
                max_CoD = 0;
				// get the currently open connections so we don't open a second connection
                nr_open_con = (u_char)bt_hci_get_con_handles(cm_stack->bt_stack, con_handles);
/*                
                bt_hci_con_handle_t hci_con_handle;
                // TODO add error check
                l2cap_cl_get_hci_connection_handle( con_handle, &hci_con_handle);
*/                
                for (i = 0; i < _bt_cmds_nr_inq_dev; i++) {
                    DEBUG("inquired device %d has CoD %d\n", i, _bt_cmds_inqres[i].cod);
                    if (_bt_cmds_inqres[i].cod == cm_stack->cod) {
                        // add this handle to possible candidates for new connection    
                        // if we're not already connected to it
						for (j = 0; j < nr_open_con; ++j) {
							if (!bt_hci_get_remote_bt_addr(cm_stack->bt_stack, con_handles[j], addr) && BD_ADDR_CMP( addr, _bt_cmds_inqres[i].bdaddr))
						       break;
						}
						if (j == nr_open_con) {
                          _right_cod_dev_indices[_right_cod_dev] = i;
                          _right_cod_dev++;
                          if (_bt_cmds_inqres[i].cod > max_CoD) {
                              max_CoD = _bt_cmds_inqres[i].cod;
                          }
						}
					}
                }
                
            }
            //check if there are possible candidates for new connections
            if (_right_cod_dev > 0) {
                DEBUG("inq (%d)\n", _right_cod_dev);
                _chosen_device = (int)(((float)_right_cod_dev) * rand() / (RAND_MAX + 1.0));
                while (_bt_cmds_inqres[_right_cod_dev_indices[_chosen_device]].cod != max_CoD) {
                    DEBUG("CMI: chosen device %d has CoD %d < %d = max_CoD\n",
                           _chosen_device, _bt_cmds_inqres[_right_cod_dev_indices[_chosen_device]].cod, max_CoD);
                    _chosen_device = (int)(((float)_right_cod_dev) * rand() / (RAND_MAX + 1.0));
                }
                    connect_result = bt_hci_create_connection(
                        cm_stack->bt_stack, BT_HCI_SYNC, 
                        _bt_cmds_inqres[_right_cod_dev_indices[_chosen_device]].bdaddr,
                        BT_HCI_PACKET_TYPE_DM1 | BT_HCI_PACKET_TYPE_DH1 | BT_HCI_PACKET_TYPE_DM3,
                        _bt_cmds_inqres[_chosen_device].page_scan_rep_mode,
                        _bt_cmds_inqres[_chosen_device].clock_offset,
                        BT_HCI_ROLE_SWITCH_CHANGE);

                INFO("connection to "ADDR_FMT", rc 0x%x\n",
                        ADDR(_bt_cmds_inqres[_right_cod_dev_indices[_chosen_device]].bdaddr),
                        connect_result);
                _right_cod_dev = 0;
            }
            // make us visible for inquiries and connects
            //bt_hci_write_scan_enable(cm_stack->bt_stack, BT_HCI_SYNC, BT_HCI_SCAN_MODE_INQUIRY_PAGE);
        }
        NutSleep(1000);
    }                           // end for       
}

THREAD(CON_EVENT_BUF, arg)
{
	NutThreadSetPriority(80);
	while (1) {
		con_event_listen();
	}
}

unsigned int _RSHash(u_char *str, unsigned int len)
{

    unsigned long int b = 378551;
    unsigned long int a = 63689;
    unsigned int hash = 0;
    unsigned int i = 0;

    for (i = 0; i < len; str++, i++) {
        hash = hash * a + (*str);
        a = a * b;
    }

    return (hash & 0x7FFFFFFF);

}

void _cm_cmds_print_usage(void)
{
    tprintf("cm: error: usage: cm id\n");
    tprintf("                  cm relcons\n");
    tprintf("                  cm inq\n");
    tprintf("                  cm autinq [ 0|1 [0|1] ]\n");
    tprintf("                  cm aiperiod [ SECONDS [0|1] ]\n");
    tprintf("                  cm aitime [ SECONDS [0|1] ]\n");
    tprintf("                  cm maxsl [1-7]\n");
    tprintf("                  cm debug [0|1]\n");
    tprintf("                  cm blink\n" );
    tprintf("                  cm disconall\n");
}

void _cm_cmd_aiperiod(char * arg)
{
    u_char len;
    unsigned int val = 0;
    unsigned int val2 = 0;

    if (sscanf(arg, "aiperiod%u%u", &val, &val2)) {
        cm_stack->sleep_period = (u_char) val;
        if (val2 == 1) {
        	++val;
            len = btn_eepromdb_store(FOURCC('A', 'I', 'P', '='), sizeof(int), &val);
        }
    } else {
        len = btn_eepromdb_get(FOURCC('A', 'I', 'P', '='), sizeof(int), &val);
        tprintf("Autinq period is %u seconds. Stored in EEPROM: %u\n",
              cm_stack->sleep_period, len == 0 ? 0 : (u_char)--val);
      }
}

void _cm_cmd_inq(char * arg)
{
    int val = 0;

    if (sscanf(arg, "inq%u", &val)) {
        cm_stack->inquiries_since_discon =val;
    } else {
        tprintf("Inq since last discon: %u\n",
              cm_stack->inquiries_since_discon);
    }
}

void _cm_cmd_autinq(char * arg)
{
    u_char len;
    unsigned int val = 0;
    unsigned int val2 = 0;
//    unsigned int val3 = 0;
//    unsigned int val4 = 0;
    
    if (sscanf(arg, "autinq%u%u", &val, &val2) && (val == 0 || val == 1)) {
        if(val){
            val = CM_AUTINQ_ON;
        }
        else {
            val = CM_AUTINQ_OFF;
        }
        cm_stack->aut_inquiry = val;
        if (val2 == 1) {
        	len = btn_eepromdb_store(FOURCC('A', 'I', 'N', '='), sizeof(unsigned int), &val);
//            DEBUG("Stored %u bytes.\n",len);
        }
    } else {
        len = btn_eepromdb_get(FOURCC('A', 'I', 'N', '='), sizeof(unsigned int), &val);
        tprintf("Autinq is %u. Stored in EEPROM: %u\n",
              cm_stack->aut_inquiry == CM_AUTINQ_ON ? 1 : 0,
              len == 0 ? 0 : (u_char) val);
    }
}

void _cm_cmd_aitime(char* arg){
	u_char len;
    unsigned int val = 0;
    unsigned int val2 = 0;

    if (sscanf(arg, "aitime%u%u", &val, &val2)) {
        cm_stack->ai_time = (u_char) val;
        if (val2 == 1) {
            ++val;
            len = btn_eepromdb_store(FOURCC('A', 'I', 'T', '='), sizeof(int), &val);
        }
    } else {
        len = btn_eepromdb_get(FOURCC('A', 'I', 'T', '='), sizeof(int), &val);
        tprintf("Autinq length is %u seconds. Stored in EEPROM: %u\n",
              cm_stack->ai_time, len == 0 ? 0 : (u_char) --val);
      }
}

void _cm_cmd_maxsl(char* arg){
	int val = 0;
	u_char len;
	if (sscanf(arg, "maxsl%u", &val)){
		cm_stack->max_slaves = (u_char) val;
		len = btn_eepromdb_store(FOURCC('C','M','S','='), sizeof(u_char), &val);	
	}
	else {
		len = btn_eepromdb_get(FOURCC('C','M','S','='), sizeof(u_char), &val);
	}
	DEBUG("max slaves = %d\n", val);
}

void _cm_cmds_cm_cmd(char * arg)
{
    if (!strncmp(arg, "relcons", 7)) {
        _bt_cm_print_reliable_cons();
    } else if (!strncmp(arg, "blink", 5)) {
        btn_led_clear_pattern_queue();
        btn_led_add_pattern(BTN_LED_PATTERN_ON_OFF, 15, 2, 4);
        btn_led_add_pattern(BTN_LED_PATTERN_ON_OFF, _bt_cm_get_nr_reliable_cons(), 3, 2);
        btn_led_add_pattern(BTN_LED_PATTERN_ON_OFF, 15, 2, 4);
    } else if (!strncmp(arg, "disconall", 9)) {
        _bt_cm_disconnect_all();
    } else if (!strncmp(arg, "id", 2)) {
        _bt_cm_print_tree_ID();
    } else if (!strncmp(arg, "autinq", 6)) {
        _cm_cmd_autinq(arg);
    } else if (!strncmp(arg, "inq", 3)) {
        _cm_cmd_inq(arg);
    } else if (!strncmp(arg, "aiperiod", 8)) {
        _cm_cmd_aiperiod(arg);
    } else if (!strncmp(arg, "aitime", 6)){
    	_cm_cmd_aitime(arg);
    } else if (!strncmp(arg, "maxsl", 5)){
    	_cm_cmd_maxsl(arg);
    } else {
        _cm_cmds_print_usage();
    }
}

void cm_tree_inq_disable(void) {
    NutSemWait(&cm_stack->inq_mutex);
    DEBUG("Inquiries disabled\n");
}

void cm_tree_inq_enable(void) {
    NutSemPost(&cm_stack->inq_mutex);
    DEBUG("Inquiries reenabled\n");
}

void cm_tree_register_cmds(void)
{
    btn_terminal_register_cmd("cm", _cm_cmds_cm_cmd);
}

short cm_tree_fill_con_info(u_char* args, u_char* param){    
    u_char i;
    bt_addr_t addr;
    long role;
    char rssi;

    *args++ = cm_stack->num_of_reliable_cons;
    
    for( i = 0 ; i < BT_HCI_MAX_NUM_CON; i++) {
        if( cm_stack->reliable_cons[i] != BT_HCI_HANDLE_INVALID) {
            /* copy bd addr */
            bt_hci_con_handle_t con_handle;
            con_handle = cm_stack->reliable_cons[i];
            bt_hci_get_remote_bt_addr( cm_stack->bt_stack,
                                       con_handle,
                                       addr);
            memcpy(args, addr, 2);
            args += 2;
            /* copy role */
            role = bt_hci_local_role_discovery( cm_stack->bt_stack,
                                                con_handle);
            if (role == BT_HCI_MY_ROLE_MASTER){
                *args++ = 2; // neighbor is my slave
            } 
            else{
                *args++ = 1; // neighbor is my master
            }
            /* copy rssi */
            bt_hci_read_abs_rssi(cm_stack->bt_stack, BT_HCI_SYNC, con_handle, &rssi);
            *args++ = rssi;
        }
    }
    return (1 + cm_stack->num_of_reliable_cons*4);
}

u_char cm_tree_get_rel_cons(bt_hci_con_handle_t *rel_cons)
{
    u_char i;
    u_char n = 0;
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        if (cm_stack->reliable_cons[i] != BT_HCI_HANDLE_INVALID){
            rel_cons[n] = cm_stack->reliable_cons[i];
            n++;
        }
    }
    return n;
}

void cm_tree_init(struct btstack *bt_stack,
					bt_psm_t* psmux,
					u_short psm,
					HCI_CON_TABLE_CB_REGISTRATION,
					u_long cod)
{
    u_char i;
    long retval;
    int val=0;
    u_char len = 0;
    bt_psm_msg_bearer* msg_bearer;
    //dyn alloc tp stack
    cm_stack = NutHeapAllocClear(sizeof(cm_stack_t));
    
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        cm_stack->reliable_cons[i] = BT_HCI_HANDLE_INVALID;
    }
    cm_stack->bt_stack = bt_stack;
    cm_stack->psmux = psmux;
    cm_stack->psm = psm;
    cm_stack->cod = cod;
    bt_hci_read_bt_addr(cm_stack->bt_stack, BT_HCI_SYNC, cm_stack->my_addr);
    //set tree ID to my addr
    memcpy(cm_stack->tree_ID, cm_stack->my_addr, BD_ADDR_LEN);
    bt_hci_register_accept_con_cb(cm_stack->bt_stack, &_cm_accept_filter, (void*)cm_stack);

    // init con_event buffer
    con_event_buffer_init(bt_stack, con_table_cb_reg);
    
    // register con table change callback at con_event buffer   
    con_event_buffer_register_con_table_cb(bt_stack, _con_table_change_cb, NULL);    

    // seed random number generator with my MAC
    srand(_RSHash(cm_stack->tree_ID, BD_ADDR_LEN));
    DEBUG("seed for rand() was %d, first random numbers are %d, %d, %d, %d, %d\n",
         _RSHash(cm_stack->tree_ID, BD_ADDR_LEN), rand(), rand(), rand(), rand(), rand());

    _bt_cm_show_tree_ID();

    btn_led_heartbeat(CM_LED_HB_PAUSE, 1, CM_LED_HB_DURATION);
    //bt_hci_write_local_name(_bt_conmgr_bt_stack, BT_HCI_SYNC, "Jaws BTnode");
    bt_hci_write_local_cod(cm_stack->bt_stack, BT_HCI_SYNC, cod);

    //init automatic inquiry
    len = btn_eepromdb_get(FOURCC('A', 'I', 'N', '='), sizeof(u_char), &val);
    if (val!=CM_AUTINQ_ON && val!=CM_AUTINQ_OFF){
	    val = CM_AUTINQ;
        btn_eepromdb_store(FOURCC('A', 'I', 'N', '='), sizeof(u_char), &val);    
    }     
    cm_stack->aut_inquiry = (u_char) val;
    
    cm_stack->inquiries_since_discon = 0;
     
    len = btn_eepromdb_get(FOURCC('A', 'I', 'P', '='), sizeof(u_char), &val);
    if(len==0){
	    val = CM_AIPERIOD;	
	    ++val;	
        btn_eepromdb_store(FOURCC('A', 'I', 'P', '='), sizeof(u_char), &val);
    }        
    cm_stack->sleep_period = (u_char) --val;
    
    len = btn_eepromdb_get(FOURCC('A', 'I', 'T', '='), sizeof(u_char), &val);
    if (len==0){
	    val = CM_AITIME;	
	    ++val;	
        btn_eepromdb_store(FOURCC('A','I','T','='), sizeof(u_char), &val);
    }        
    cm_stack->ai_time = (u_char) --val;
    
    len = btn_eepromdb_get(FOURCC('C', 'M', 'S', '='), sizeof(u_char), &val);
    if (len==0){
	    val = CM_MAX_SLAVES;	
        btn_eepromdb_store(FOURCC('C','M','S','='), sizeof(u_char), &val);
    }        
    cm_stack->max_slaves = (u_char) val;
    val = CM_MAX_SLAVES;
	    

	// initialize mutex
    NutMutexInit(&cm_stack->rel_con_cb_mutex);
    NutSemInit(&cm_stack->inq_mutex, 1);

    // set default link policy (disable all)
    retval = bt_hci_write_default_link_policy_settings(cm_stack->bt_stack, BT_HCI_SYNC, 0x0000);
    DEBUG("link policy changed = %d\n", retval);

	// register cm on l2cap_cl layer
	retval =
		bt_psm_service_register(psmux, psm, _cm_data_cb, NULL);
	if (retval >= 0) {
		// create pkt buffer queue
		cm_stack->buf_que = bt_acl_pkt_queue_create(bt_stack, 3);
		msg_bearer = bt_psm_msg_bearer_create(1024, "CM_DATA");
		bt_psm_service_set_buffers(psmux, retval, &cm_stack->buf_que, msg_bearer);
	}
	else {
		ERROR("could not register data callback\n");
		return;
	}
	
	// initialize con mgr interface
	con_mgr_stack.cm_fill_con_info = cm_tree_fill_con_info;
	con_mgr_stack.cm_inq_disable = cm_tree_inq_disable;
	con_mgr_stack.cm_inq_enable = cm_tree_inq_enable;
	con_mgr_stack.cm_reg_cmds = cm_tree_register_cmds;
	con_mgr_stack.cm_get_rel_cons = cm_tree_get_rel_cons;
	
    // start conmgr thread
    NutThreadCreate("ConmgrI", CONMGR_INQUIRY_T, 0, NUT_DEFAULT_STACKSIZE);
    // start con event buffer
    NutThreadCreate("CON_BUF", CON_EVENT_BUF, 0, NUT_DEFAULT_STACKSIZE);
}
