/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

 /**
 * @file mhop_cl.c
 *
 *\author Kevin Martin <kevmarti@tik.ee.ethz.ch>
 * 
 *\date 10-14-05
 * 
 *\brief Implementation of the connection-less multi-hop layer
 *
 */
#include <string.h>
#include <sys/event.h>
#include <sys/heap.h>
#include <sys/thread.h>
#include <sys/mutex.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_hci_cmds.h>
#include <terminal/btn-terminal.h>
#include <bt/bt_psm.h>
#include <bt/l2cap_cl.h>
#include <mhop/mhop_cl.h>
#include <debug/toolbox.h>
#include <sys/timer.h>

// debug settings
#define LOG_CLASS 6
#define LOG_LEVEL 0
#include <debug/log_set.h>

#if LOG_LEVEL >= 3
u_char _mhop_cl_dbg_buf[3*121];
#endif

/// paramater for avoiding multiple reception of broadcast packets
#define MHOP_CL_BC_WINDOW 64

#define BT_HCI_CON_HANDLE_OWN 0

// flag bits
#define MHOP_CL_FLAG_BIT_SRC_DEL	7
#define MHOP_CL_FLAG_BIT_SRC_REC	6
#define MHOP_CL_FLAG_BIT_VERBOSE	5

// flag accessors
#define MHOP_CL_FLAG_SRC_DEL		0x01 << MHOP_CL_FLAG_BIT_SRC_DEL
#define MHOP_CL_FLAG_SRC_REC		0x01 << MHOP_CL_FLAG_BIT_SRC_REC
#define MHOP_CL_FLAG_VERBOSE		0x01 << MHOP_CL_FLAG_BIT_VERBOSE


/** 
 * \brief Definition of the multi-hop control header.
 * 
 * The information fields of the multi-hop control header are required for
 * routing data properly to a remote device via several intermediate hops.
 * \see mhop_cl_send_pkt_to()
 * 
 \verbatim --------------------------------------------------------------------------------
 | flags | seq | ttl | hop_cnt |  source   | destination | src_port | dest_port |
 --------------------------------------------------------------------------------
 .   1      1     1       1     BD_ADDR_LEN  BD_ADDR_LEN      1           1\endverbatim
 */
typedef struct _mhop_cl_pkt_hdr {
	u_char flags;				///< flags, defining the type of the packet
    u_char seq;                 ///< broadcast sequence number
    u_char ttl;					///< time to live (0 = inf)
    u_char hop_cnt;             ///< hop count (distance to host)
    bt_addr_t source;           ///< address of the sending node
    bt_addr_t destination;      ///< address of the destination node
    u_char psm[L2CAP_CL_PSM_FIELD_SIZE];
    							///< psm the packet will be delivered to on
    							///< the receiving side
} mhop_cl_pkt_hdr;

/**
 * \brief Forwarding table entry
 * 
 * When receiving a connectionless l2cap packet containing a multi-hop
 * information trailer whos "source record" flag is set
 * (see typedef #mhop_cl_mhop_ctrl_info), the address of the sender will
 * be stored in the forwarding table.
 * 
 * Essentially, an entry in the forwarding table consists of
 * the address of the source node, and the connection handle over which
 * the source node can be reached.
 */
typedef struct _mhop_cl_fwd_table_entry {
    bt_addr_t src_addr;			///< Address of the source node
    bt_hci_con_handle_t handle;	///< Outgoing handle
    u_char seq;                 ///< hold the last seq nr for every src node!
    u_char distance;            ///< distance to the source node
} _mhop_cl_fwd_table_entry_t;

/**
 * \brief connection-less multi-hop stack.
 */
struct _mhop_cl_stack_s {
	struct btstack* bt_stack;
    							///< Pointer to the lower layer l2cap cl stack
    bt_psm_t* psmux;			///< Pointer to the protocol / service multiplexor
	u_short mhop_psm;			///< PSM of the multi-hop cl service
	u_short fwd_psm;			///< PSM used for packet forwarding service
	bt_acl_pkt_buf* fwd_bufs;	///< Pointer to the packet buffers used for forwarding
    bt_addr_t my_addr;          ///< own bt addr
    _mhop_cl_fwd_table_entry_t fwdtable[MHOP_CL_MAX_FWD_TABLE_SIZE];
    							///< forwarding table
    u_char mylastseq;           ///< last sequence number I(!) used
    bt_hci_con_handle_t reliable_con_handles[BT_HCI_MAX_NUM_CON];
    							///< list with reliable connection handles
} _mhop_cl_stack;


/**
 * \brief builds the multi-hop packet header.
 * 
 * \param mhop_ctrl Pointer to the trailer
 * \param srd_del Source deletion flag
 * \param src_ref Source recording flag
 * \param verbose Verbosity flag
 * \param ttl Time to live of the packet
 * \param seq Broadcast sequence number
 * \param source BT address of the sender of this packet
 * \param destination BT address of the receiver
 * \param psm PSM this packet shall be sent to
 */
void _mhop_cl_build_pkt_hdr(mhop_cl_pkt_hdr* mhop_hdr,
	u_char src_del, u_char src_rec, u_char verbose, u_char ttl, u_char seq,
	bt_addr_t source, bt_addr_t destination, u_short psm)
{
    // build 'flags' field
    mhop_hdr->flags = ((src_del << MHOP_CL_FLAG_BIT_SRC_DEL) |
					(src_rec << MHOP_CL_FLAG_BIT_SRC_REC) |
    				(verbose << MHOP_CL_FLAG_BIT_VERBOSE));
    // set broadcast stuff
    mhop_hdr->seq = seq;
    mhop_hdr->ttl = ttl;
    mhop_hdr->hop_cnt = 0;
    // set source and destination addresses
    memcpy(mhop_hdr->source, source, BD_ADDR_LEN);
    memcpy(mhop_hdr->destination, destination, BD_ADDR_LEN);
    // destination port
    mhop_hdr->psm[0] = BTN_LO(psm);
    mhop_hdr->psm[1] = BTN_HI(psm);
}

/**
 * Returns the conneciton handle of the desired destination \p dest.
 * 
 * Scans the forwarding table for an entry containing the destination
 * address \p dest, and returns the corresponding connection handle.
 * 
 * \param stack Pointer to the connectionless l2cap stack
 * \param dest Address of the desired destination node
 * 
 * \return
 * - the corresponding out-handle if there is an entry
 * - #BT_HCI_HANDLE_INVALID else
 */
bt_hci_con_handle_t _mhop_cl_ft_get_handle_of(bt_addr_t dest)
{
    u_short i;
    for (i = 0; i < MHOP_CL_MAX_FWD_TABLE_SIZE; i++)
    {
        if (BD_ADDR_CMP(_mhop_cl_stack.fwdtable[i].src_addr, dest))
            return _mhop_cl_stack.fwdtable[i].handle;
    }
    DEBUG("no out_handle found!\n");
    return BT_HCI_HANDLE_INVALID;
}


/**
 * \brief Updates the forwarding table.
 *
 * \param src_addr Address of the source
 * \param src_handle Handle pkt was received from
 * \param src_seq Sequence number of the source broadcast
 * \param ttl time to live of the source broadcast
 * \param distance Distance to the source
 */
u_char _mhop_cl_update_fwd_table(bt_addr_t src_addr,
			bt_hci_con_handle_t src_handle, u_char src_seq, u_char distance) {
    u_short i = 0;
    u_short j;
    
    // find empty or same entry in fwd table
    j = MHOP_CL_MAX_FWD_TABLE_SIZE;
    for (i = 0; i < MHOP_CL_MAX_FWD_TABLE_SIZE; i++)
    {
        if ((BD_ADDR_CMP(_mhop_cl_stack.fwdtable[i].src_addr, src_addr)) &&
        	(_mhop_cl_stack.fwdtable[i].handle != BT_HCI_HANDLE_INVALID))
        	break;
        else if (j == MHOP_CL_MAX_FWD_TABLE_SIZE && _mhop_cl_stack.fwdtable[i].handle == BT_HCI_HANDLE_INVALID)
        {
            j = i;      // remember first empty entry
        }
    }
    // already received this bc?
    if (i == MHOP_CL_MAX_FWD_TABLE_SIZE && j == MHOP_CL_MAX_FWD_TABLE_SIZE)
    {
        ERROR("fwd table full!\n");
        return 1;
    }
    else if (i == MHOP_CL_MAX_FWD_TABLE_SIZE)
    {
        DEBUG("new source stored.\n");
        //new source
        //add incoming connection to fwd table
        memcpy(_mhop_cl_stack.fwdtable[j].src_addr, src_addr, BD_ADDR_LEN);
        _mhop_cl_stack.fwdtable[j].seq = src_seq;
        _mhop_cl_stack.fwdtable[j].handle = src_handle;
        _mhop_cl_stack.fwdtable[j].distance = distance;

    }
    // source was in table 
    // check seq nr! if seq nr is in broadcast window store handle and proceed
    else if (((_mhop_cl_stack.fwdtable[i].seq - src_seq) % (sizeof(src_seq)*256)) <= MHOP_CL_BC_WINDOW)
    {
        // bc allready received
        // do nothing
        return 3;
    }
    else
    {
    	// update fwdtable
        _mhop_cl_stack.fwdtable[i].handle = src_handle;
        _mhop_cl_stack.fwdtable[i].seq = src_seq;
        _mhop_cl_stack.fwdtable[i].distance = distance;
       	DEBUG("from host "ADDR_FMT" with seq nr %d over handle %d\n",
                ADDR(src_addr),
                src_seq, src_handle);
    }
    return 0;
}


long _mhop_cl_send_pkt_to(mhop_cl_pkt_hdr* hdr,
							u_char* data,
							u_short len,
							bt_hci_con_handle_t hdl)
{
	long res;
	l2cap_cl_channel_t* cl_chan;
	
	// get access to the l2cap cl channel 
    
	res = l2cap_cl_channel_open(hdl,
								_mhop_cl_stack.mhop_psm,
								len + sizeof(mhop_cl_pkt_hdr),
								&cl_chan);							
	if (res) {
		WARNING("opening cl chan: %d\n", res);
		return res;
	}
	
	// write mhop header
	res = l2cap_cl_channel_write(cl_chan, (u_char*)hdr, sizeof(mhop_cl_pkt_hdr));
	if (res) {
		WARNING("writing hdr: %d\n", res);
		return res;
	}
	// write data
	if (len) {
		res += l2cap_cl_channel_write(cl_chan, data, len);
	}
	if (res) {
		WARNING("writing data: %d\n", res);
		return res;
	}
					
	// close the channel
	l2cap_cl_channel_close(cl_chan);
	return 0;
}

/**
 * \brief Forwards a packet to the target device.
 * 
 * If an entry for the target device exists in the forwarding table (that is
 * we know how to reach the target device), the packet is forwarded to the
 * corresponding next hop.
 * 
 * If there is no entry in the forwarding table, the packet is broadcasted.
 * 
 * \param stack Pointer to the connectionless multi-hop stack
 * \param hdr Pointer to the mhop cl header
 * \param data Pointer to the data to be sent
 * \param len Length of the data to be sent (excluding mhop cl header)
 * \param in_handle Connection handle the packet was received on.
 */
long _mhop_cl_fwd_pkt(mhop_cl_pkt_hdr* hdr, u_char* data,
								u_short len, bt_hci_con_handle_t in_handle)
{
	u_short i;
	long res;
	bt_hci_con_handle_t* rel_cons;
    bt_hci_con_handle_t out_handle;
    
	// increment hop count
	hdr->hop_cnt++;
	// Check if the desired destination can be found in the forwarding table
	out_handle =
		_mhop_cl_ft_get_handle_of(hdr->destination);
	if (out_handle != BT_HCI_HANDLE_INVALID) {
        // forward
       	DEBUG("forwarding pkt, out_handle: %d\n", out_handle);
        return _mhop_cl_send_pkt_to(hdr, data, len, out_handle);
	}
	// else, forward the packet to all reliable connections except to the
	// con the packet was received on
	DEBUG("broadcasting pkt...\n");
	rel_cons = _mhop_cl_stack.reliable_con_handles;
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++)
    {
        if ((rel_cons[i] != BT_HCI_HANDLE_INVALID) && (rel_cons[i] != in_handle) 
            						&& (rel_cons[i] != BT_HCI_CON_HANDLE_OWN))
        {
            // send packet to all connections in reliable connections
            DEBUG("bc: sending to handle: %d...\n", rel_cons[i]);
            									
            res = _mhop_cl_send_pkt_to(hdr, data, len, rel_cons[i]);
            if (res)
            {
                DEBUG("bc: error! hdl %d (0x%.2x)\n", rel_cons[i], -res);
            }
            else
            	DEBUG("bc to hdl %d sent\n", rel_cons[i]);
        }
    }
	return 0;
}
    
/**
 * Is called each time a packet has to be forwarded.
 * 
 * Since forwarding may require some time (e.g. if we have to broadcast the
 * packet), received packets are forwarded within the context of a special
 * thread. The required context-switch is implemented by registering a
 * forwarding service at the protocol / service multiplexor which provides
 * the buffering / context-switch needed "for free".
 * 
 * This is the callback function registered at the protocol/service multiplexor
 * during initialization.
 * 
 * \see #bt_psm.h
 * \see #BT_PSM_DATA_CB
 * \see #mhop_cl_init
 */
bt_msg* _mhop_cl_fwd_queue_data_cb(bt_msg* msg,
									u_char* data,
									u_short len,
									u_short service_nr,
									void* cb_arg)
{
    bt_hci_con_handle_t in_handle;
    mhop_cl_pkt_hdr* hdr;
    bt_msg* free_msg;
    struct bt_hci_pkt_acl* pkt;
    u_short psm;
    long retval;
    
    // get pointer to the buffered packet
    pkt = msg->pkt;
    
    // init pointer to the next free pkt buffer
    free_msg = msg;
    
    // get pointer to the multi-hop header
    hdr = (mhop_cl_pkt_hdr*)(data - sizeof(mhop_cl_pkt_hdr));
    
	// get con handle the packet was received on
	in_handle = bt_acl_get_con_handle(pkt);
	
	// forward the packet
	retval = _mhop_cl_fwd_pkt(hdr, data, len, in_handle);
	if (retval) INFO("error fwd pkt: %d\n", retval);
	
    // if verbose flag is set or destination addr is the broadcast addr, we
    // additionally have to deliver the packet
    if (BD_ADDR_CMP(hdr->destination, bt_addr_null) || (hdr->flags & MHOP_CL_FLAG_VERBOSE))
    {
    	psm = hdr->psm[0] | hdr->psm[1] << 8;
        if(psm != 0){
            retval =
        		bt_psm_deliver_msg(_mhop_cl_stack.psmux, msg, data, len, psm, &free_msg);
    	   if (retval) WARNING("couldn't deliver pkt to psm %d! code: %d\n", psm, retval);
        }
    }
	return free_msg;
}

/**
 * \brief l2cap cl data callback.
 * 
 * This callback is called each time an l2cap cl packet arrives on the
 * psm #mhop_cl_stack::mhop_psm. This callback is registered at the
 * protocol / service multiplexor during initalization of the
 * connection-less multi-hop layer.
 */
bt_msg* _mhop_cl_data_cb(bt_msg* msg,
							u_char* data,
							u_short len,
							u_short service_nr,
							void* cb_arg)
{
	u_short i;
	u_short psm;
	mhop_cl_pkt_hdr* mhop_hdr;
	u_char flags;
	bt_hci_con_handle_t in_handle;
	u_char* source;
	u_char* target;
	long retval;
	bt_msg* next_msg;
	struct bt_hci_pkt_acl* pkt;
	
	next_msg = msg;

	// get pointer to the packet
	pkt = msg->pkt;
	// get pointer to multi-hop packet header
	mhop_hdr = (mhop_cl_pkt_hdr*)data;
	// adjust payload pointer
	data += sizeof(mhop_cl_pkt_hdr);
	// adjust payload size
	len -= sizeof(mhop_cl_pkt_hdr);
	// get psm the packet has to be delivered to
	psm = mhop_hdr->psm[0] | mhop_hdr->psm[1] << 8;
    // get flags
    flags = mhop_hdr->flags;
    // get address of target device & sending device
    source = mhop_hdr->source;
    target = mhop_hdr->destination;
    // get connection handle the packet was received on
    in_handle = bt_acl_get_con_handle(pkt);

	/*
	 * con handle check
	 */
	if (!(in_handle == BT_HCI_CON_HANDLE_OWN))
	{
		// check if con handle is reliable
	    for( i = 0; i < BT_HCI_MAX_NUM_CON; i++)
	    {
	        if (_mhop_cl_stack.reliable_con_handles[i] == in_handle) break;
	    }
	    if( i == BT_HCI_MAX_NUM_CON)
	    {
	        WARNING("no con handle to this pkt!!\n");
	        // disconnect ?
	        return msg;
	        //btn_led_clear_pattern_queue();
	        //btn_led_add_pattern(BTN_LED_PATTERN_ON_OFF, 11, 1, BTN_LED_INFINITE);
	    }
	    //TODO: check if this con_handle is already in reliable con table, if not, wait for this
	    //reliable connection. this problem occures at new connection. other deivce sends
	    //the host entry pkt (if host entries are found) before the reliable con changed is called!!
	}
	
	/*
	 * debugging info
	 */
	DEBUG("rcvd mhop pkt: %d bytes for psm %d, handle: %d src: "ADDR_FMT"\n",
					len, psm, in_handle, ADDR(source));
	DEBUG("dest: "ADDR_FMT", flags: %.2x\n", ADDR(target), flags);
    DEBUG_STR("%s\n", dbg_str_to_hex_str(_mhop_cl_dbg_buf, data, len));

	/*
	 * a source deletion packet arrived !
	 * 
	 * if the packet arrived on the same handle as in the forwarding table,
	 * delete the source
	 */
	if (flags & MHOP_CL_FLAG_SRC_DEL)
	{
        DEBUG("src del pkt rcvd! handle: %d, host: "ADDR_FMT"\n",
                 in_handle,
                 ADDR(source));
        //find source and delete it
        for (i = 0; i < MHOP_CL_MAX_FWD_TABLE_SIZE; i++)
        {
        	// skip invalid entries
            if (_mhop_cl_stack.fwdtable[i].handle != BT_HCI_HANDLE_INVALID)
            {
            	// find source
                if (BD_ADDR_CMP(_mhop_cl_stack.fwdtable[i].src_addr, source))
                {
                	// only delete, if source direction was on this con
		            if (_mhop_cl_stack.fwdtable[i].handle == in_handle)
		            {
		                // source really lost, delete it from forwarding table
		                // and forward the packet
		                _mhop_cl_stack.fwdtable[i].handle = BT_HCI_HANDLE_INVALID;
		                _mhop_cl_stack.fwdtable[i].seq = 0;
                        DEBUG("src deleted.\n");
						// forward pkt by adding it to the forwarding queue
						bt_psm_deliver_msg(_mhop_cl_stack.psmux,
		                					msg,
		                					data,
		                					len,
		                					_mhop_cl_stack.fwd_psm,
		                					&next_msg);
		            }
		            break;
                }
            }
        }
	    return next_msg;
	}
	
	/*
	 * source recording flag is set
	 * 
	 * if the source of the packet should be recorded (added to the forwarding
	 * table), update the forwarding table accordingly
	 */
	else if (flags & MHOP_CL_FLAG_SRC_REC)
	{
	    DEBUG("src rec pkt rcvd!\n");
	    retval = _mhop_cl_update_fwd_table(source,
	    									in_handle,
	    									mhop_hdr->seq,
	    									mhop_hdr->hop_cnt);
	    if (retval == 1)
	    {
	        WARNING("ft error\n");
	        return msg;
	    }
	    else if (retval == 2)
	    {
	        WARNING("ft entry deleted\n");
	        return msg;
	    }
	    else if (retval == 3)
	    {
	        WARNING("pkt already rcvd\n");
	        return msg;
	    }
        // else if retval == 0 (normal case)
	}
	
	/*
	 * Packet delivery & forwarding
	 * 
	 * A packet is delivered directly to the service only if the target address
	 * equals the address of this device.
	 * If this is not the case, the packet is added to the forwarding queue iff
	 * the ttl of the packet is still ok.
	 */
	if (!BD_ADDR_CMP(target, _mhop_cl_stack.my_addr)) {
		if (mhop_hdr->hop_cnt <= mhop_hdr->ttl || mhop_hdr->ttl == 0){
			psm = _mhop_cl_stack.fwd_psm;
		}
		else return next_msg;
	}
	// deliver the packet
    DEBUG("deliver pkt\n");
	retval = bt_psm_deliver_msg(_mhop_cl_stack.psmux,
								msg,
								data,
								len,
								psm,
								&next_msg);
	// debugging info
	if (retval) INFO("couldn't deliver pkt to psm %d! code: %d\n", psm, retval);
	return next_msg;
}

// Callback function for changes in reliable connections
void _mhop_cl_reliable_con_change_cb(u_char type, u_char detail, bt_hci_con_handle_t con_handle, void* cb_arg)
{
    u_char i;
    bt_hci_con_handle_t* rel_cons;
    _mhop_cl_fwd_table_entry_t* fwt;
    mhop_cl_pkt_hdr pkt_hdr;
    
    rel_cons = _mhop_cl_stack.reliable_con_handles;
    if (type == BT_HCI_CONNECTION) {
        DEBUG("new con! detail: %d, handle: %d\n",
                      type, detail, con_handle);
        // add handle to reliable connections table
        for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
            if (rel_cons[i] == con_handle) // add a con_handle only once!
                break;
            if (rel_cons[i] == BT_HCI_HANDLE_INVALID) {
                rel_cons[i] = con_handle;
                break;
            }
        }
    } else if (type == BT_HCI_DISCONNECTION) {
        DEBUG("disconnect! detail: %d, handle: %d\n",
                      type, detail, con_handle);
        // remove handle from reliable connections table
        for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
            if (rel_cons[i] == con_handle)
                rel_cons[i] = BT_HCI_HANDLE_INVALID;
        }
        // check if connection to a known source broke!
        for (i = 0; i < MHOP_CL_MAX_FWD_TABLE_SIZE; i++) {
            fwt = &_mhop_cl_stack.fwdtable[i];
            if (fwt->handle == con_handle) {
                DEBUG("was conn to known src ("ADDR_FMT")!\n", ADDR(fwt->src_addr));                
                //connection to source lost!
                fwt->handle = BT_HCI_HANDLE_INVALID;
                fwt->seq = 0;
                //send out host del pkt!
                _mhop_cl_build_pkt_hdr(&pkt_hdr, 
                                       1, 0, 0, 0,  // flags: src del
                                       _mhop_cl_stack.mylastseq++, 
                                       fwt->src_addr, 
                                       bt_addr_null, // not used
                                       0);           // not used
                _mhop_cl_fwd_pkt(&pkt_hdr, NULL, 0, BT_HCI_HANDLE_INVALID);
            }
        }
    }
    else{   
        DEBUG("rel-con change: handle = %d, type = %d\n", con_handle, type);
    }
}

// send multi-hop packet
long mhop_cl_send_pkt(u_char* data,
						u_short len,
						u_char src_rec,
						u_char verbose,
						u_char ttl,
						bt_addr_t dest,
						u_short psm)
{
	mhop_cl_pkt_hdr hdr;

	// return if payload size is too big
	if (len > BT_ACL_PKT_MAX_PAYLOAD(_mhop_cl_stack.bt_stack) - L2CAP_CL_PKT_HDR_LEN - sizeof(mhop_cl_pkt_hdr))
		return MHOP_CL_ERR_PAYLOAD_OVERFLOW;

	// build multi-hop header
    _mhop_cl_build_pkt_hdr(&hdr, 0, src_rec, verbose, ttl,
    						_mhop_cl_stack.mylastseq++, _mhop_cl_stack.my_addr, dest, psm);

    // send (forward the packet)
    DEBUG("sending mhop pkt with seq %d\n", _mhop_cl_stack.mylastseq - 1);
    if (BD_ADDR_CMP(dest, _mhop_cl_stack.my_addr) || BD_ADDR_CMP(dest, bt_addr_null))
    	return _mhop_cl_send_pkt_to(&hdr, data, len, BT_HCI_CON_HANDLE_OWN);
    else {
    	return _mhop_cl_fwd_pkt(&hdr, data, len, BT_HCI_CON_HANDLE_OWN);
    }
}

long mhop_cl_send_src_del_pkt(u_char  ttl)
{
	mhop_cl_pkt_hdr hdr;
	
	// build multi-hop header
	_mhop_cl_build_pkt_hdr(&hdr, 1, 0, 0, ttl, _mhop_cl_stack.mylastseq++,
										_mhop_cl_stack.my_addr, bt_addr_null, 0);
		
    // broadcast the packet
    DEBUG("sending src del pkt with seq %d\n", _mhop_cl_stack.mylastseq - 1);
	return _mhop_cl_fwd_pkt(&hdr, NULL, 0, BT_HCI_CON_HANDLE_OWN);
}


u_char* mhop_cl_get_target_addr(struct bt_hci_pkt_acl* pkt)
{
	return ((mhop_cl_pkt_hdr*)l2cap_cl_get_data_ptr(pkt))->destination;
}

u_char* mhop_cl_get_source_addr(struct bt_hci_pkt_acl* pkt)
{
	return ((mhop_cl_pkt_hdr*)l2cap_cl_get_data_ptr(pkt))->source;
}

u_short mhop_cl_max_payload_size(void)
{
	return l2cap_cl_max_payload() - sizeof(mhop_cl_pkt_hdr);
}

u_char* mhop_cl_get_my_addr(void)
{
	return _mhop_cl_stack.my_addr;
}

void mhop_cl_init(struct btstack* bt_stack,
					bt_psm_t* psmux,
					u_short mhop_psm,
					u_short fwd_psm,
					u_short buf_size,
					HCI_CON_TABLE_CB_REGISTRATION)
{
	u_short i;
	long service_nr;
	bt_psm_msg_bearer* fwd_td;
	
	// store pointer to bt-stack
	_mhop_cl_stack.bt_stack = bt_stack,
	// store pointer to protocol / service multiplexor
	_mhop_cl_stack.psmux = psmux;
	// store mhop psm to use
	_mhop_cl_stack.mhop_psm = mhop_psm;
	// store packet forwarding psm
	_mhop_cl_stack.fwd_psm = fwd_psm;
    // initialize forwarding table
    for (i = 0; i < MHOP_CL_MAX_FWD_TABLE_SIZE; i++) {
        _mhop_cl_stack.fwdtable[i].handle = BT_HCI_HANDLE_INVALID;
        _mhop_cl_stack.fwdtable[i].seq = 0;
    }
    for (i = 0; i < BT_HCI_MAX_NUM_CON; i++) {
        _mhop_cl_stack.reliable_con_handles[i] = BT_HCI_HANDLE_INVALID;
    }
    // init broadcasting seq. nr.
    _mhop_cl_stack.mylastseq = 1;

    // Read own bt address
    bt_hci_read_bt_addr(_mhop_cl_stack.bt_stack, BT_HCI_SYNC, _mhop_cl_stack.my_addr);
    
	// register multi-hop service at protocol / service multiplexor
	service_nr =
		bt_psm_service_register(psmux, mhop_psm, _mhop_cl_data_cb, NULL);
	// we don't need any buffers for the mhop protocol
	bt_psm_service_set_buffers(psmux, service_nr, NULL, NULL);
	
	// initiate a service to enable buffered packet forwarding
	service_nr =
		bt_psm_service_register(psmux, fwd_psm, _mhop_cl_fwd_queue_data_cb, NULL);
	_mhop_cl_stack.fwd_bufs = bt_acl_pkt_queue_create(bt_stack, buf_size);
	fwd_td = bt_psm_msg_bearer_create(512, "mhop_fwd");
	bt_psm_service_set_buffers(psmux, service_nr, &_mhop_cl_stack.fwd_bufs, fwd_td);

	// register con change callback
	con_table_cb_reg(bt_stack, _mhop_cl_reliable_con_change_cb, NULL);
}
