#include <sys/heap.h>
#include <mhop/rpc.h> 
#include <string.h>
#include <debug/logging.h>
#include <debug/toolbox.h>
#include <terminal/btn-terminal.h>
#include <sys/timer.h>

#define LOG_LEVEL 0
#define LOG_CLASS 9
#include <debug/log_set.h>

u_char _rpc_dbg_buf[3*121];

/**
 * \brief registered PSM callback for rpc data.
 * Finds the corresponding rpc procedure, calls it and sends results back to
 * source.
 */
bt_msg* _rpc_proc_cb(bt_msg* msg, 
                       u_char* data,
                       u_short data_len, 
                       u_short service_nr, 
                       void* cb_arg)
{
    // process pkt
    u_char* source_addr;
    u_char more_pkts;
    u_char result_len = 0;
    
    DEBUG("proc cb\n");
    source_addr = mhop_cl_get_source_addr(msg->pkt);
    u_char proc_nr = data[0];
    if(proc_nr >= _rpc->max_procs){ 
        return msg;
    }
    _rpc_procs_t* proc = &_rpc->procs[proc_nr];
    // call rpc procedure
    do{
        DEBUG("call proc nr %d\n", proc_nr);
        more_pkts = proc->rpc_proc((char *)data+1, data_len-1, proc->proc_arg,
                                   _rpc->result_data+1, &result_len);
        // send result
        _rpc->result_data[0] = proc_nr;  
        if(result_len > 0){
            DEBUG("sending result to "ADDR_FMT" len= %d\n",
                      ADDR(source_addr), 
                      result_len+1);
            DEBUG_STR("%s\n", 
                      dbg_str_to_hex_str(_rpc_dbg_buf, _rpc->result_data, result_len+1));
            mhop_cl_send_pkt(_rpc->result_data, 
                             result_len + 1, 
                             0, 
                             0, 
                             0, 
                             source_addr, 
                             _rpc->result_psm);
        }
    } while (more_pkts > 0);
    // free packet buffer
    return msg;
}

/**
 * \brief registered PSM callback for handling returned rpc results.
 */
bt_msg* _rpc_result_cb(bt_msg* msg, 
                       u_char* data,
                       u_short data_len, 
                       u_short service_nr, 
                       void* cb_arg)
{
    // process pkt
    DEBUG("result cb for proc %d\n", data[0]);
    u_char* source_addr = mhop_cl_get_source_addr(msg->pkt);
    u_char proc_nr = data[0];
    if(proc_nr >= _rpc->max_procs){
        return msg;
    }
    _rpc_procs_t* proc = &_rpc->procs[proc_nr];
    u_char* result = data+1; 
    // call result procedure
    
    DEBUG("call result\n");
    proc->rpc_result(source_addr, result, data_len-1, proc->result_arg);
    // free packet buffer
    return msg;
}


void rpc_send_query(u_char proc_nr,  
                    u_char* data, u_char data_len,
                    bt_addr_t dest_addr, u_char ttl)
{                    
    u_char pkt[data_len + 1];
    pkt[0] = proc_nr;
    memcpy(pkt+1, data, data_len);
    // send query
    DEBUG("sending query to "ADDR_FMT"\n", ADDR(dest_addr));
    mhop_cl_send_pkt(pkt, 
                     data_len + 1, 
                     1, 
                     0, 
                     ttl, 
                     dest_addr, 
                     _rpc->proc_psm);
}

// TODO mutex protect _rpc->result_data
void rpc_local_query(u_char proc_nr, u_char* data, u_char data_len)
{
    u_char more_pkts;
    u_char result_len = 0;
    DEBUG("local query proc %d\n",proc_nr);
    _rpc_procs_t* proc = &_rpc->procs[proc_nr];
    // call rpc procedure
    do{
        more_pkts = proc->rpc_proc((char*)data, data_len, proc->proc_arg,
                                          _rpc->result_data+1, &result_len);
        if(result_len > 0){
            proc->rpc_result(mhop_cl_get_my_addr(), _rpc->result_data+1, result_len, proc->result_arg);
        }
    } while(more_pkts > 0);
}                     


char rpc_register_proc(u_char* proc_nr, \
                       RPC_PROC, void* proc_arg, \
                       RPC_RESULT, void* result_arg)
{
    u_char i;
    _rpc_procs_t* proc;
    for( i=0; i < _rpc->max_procs; i++){
        proc = &_rpc->procs[i];
        if(proc->rpc_proc == NULL){
            // insert service
            proc->rpc_proc = rpc_proc;
            proc->proc_arg = proc_arg;
            proc->rpc_result = rpc_result;
            proc->result_arg = result_arg;
            *proc_nr = i;
            break;      
        }
    }
    if(i == _rpc->max_procs){
        return RPC_ERR_TOO_MANY_PROCS;
    }
    else return 0;    
}                     

inline u_short rpc_max_result_size(void){
    return mhop_cl_max_payload_size() - 1;
}


rpc_stack_t* rpc_init(bt_psm_t* psmux,
						u_char nr_procs,
						u_short proc_psm,
						u_short result_psm)
{
	long service_nr;
	bt_psm_msg_bearer* bearer;
	
    _rpc = NutHeapAllocClear(sizeof(rpc_stack_t));
    _rpc->procs = NutHeapAllocClear(nr_procs * sizeof(_rpc_procs_t));
    _rpc->max_procs = nr_procs;
    _rpc->proc_psm = proc_psm;
    _rpc->result_psm = result_psm;
    _rpc->result_data = NutHeapAllocClear(mhop_cl_max_payload_size());
    
    
    // register rpc service
    service_nr = bt_psm_service_register(psmux, proc_psm, _rpc_proc_cb, NULL);
    bearer = bt_psm_msg_bearer_create(512, "rpc_proc");
    bt_psm_service_set_buffers(psmux, service_nr, bt_psm_get_default_bufs(psmux), bearer);
    bt_psm_service_register(psmux, result_psm, _rpc_result_cb, NULL);
       
    return _rpc;
}


